//-----------------------------------------------------------------------------------
//        GM-MGMC SAMPLER FOR OSV AND OSVt MODEL --- by GERNOT MUELLER, 2007
//-----------------------------------------------------------------------------------
// Change program only in the marked section below !
//
// Steps needed before executing the program:
// - adapt the program code to your data and files (only in marked section!)
// - save the program (as osvt.cpp)
// - compile the program (by 'CC -O4 -c osvt.cpp')
// - link it with all o-files (CC -O4 -o osvt osvt.o randomGM.o matrixGM.o nelderGM.o)
// - type osvt.exe to execute
//
// Maximal number of X-covariates:   10 (including intercept)
// Maximal number of Z-covariates:    6
// Maximal number of cutpoints:       6
// Maximal length of data:        60000
//-----------------------------------------------------------------------------------

#include<stdio.h>
#include<stdlib.h>
#include<iostream.h>
#include<math.h>

#include"shared1.h"
#include"shared2.h"
#include"matrixGM.h"
#include"randomGM.h"
#include"func.h"
#include"nelderGM.h"

//---------------------------------------------------------------------------------//
// function declarations
		
void simu(double mu, double phi, double sigma1);
double impactX(void);
void draw_CP(void);
void draw_Y(void);
void draw_s(void);
void draw_nu(void);
void draw_LAMBDA(void);
void draw_H(void);
void draw_BETA(void);
void Hesse(unsigned int n, double griddist=0.0001, double x1=0.00000, 
           double x2=0.00000, double x3=0.00000, double x4=0.00000, 
	   double x5=0.00000, double x6=0.00000, double x7=0.00000, 
	   double x8=0.00000, double x9=0.00000, double x10=0.00000);
double minimum(double x, double y);

//---------------------------------------------------------------------------------//
//---------------------------------------------------------------------------------//
//                                                                                 //
//                C H A N G E    O N L Y    T H I S    S E C T I O N               //
//                                                                                 //
//---------------------------------------------------------------------------------//
//---------------------------------------------------------------------------------//
//                             DATA AND MODEL PARAMETERS                           //
//------------------------------------------------ --------------------------------//
#define TTTT 20000         // length of data                                       //
#define NNNN 4             // number of parameters to estimate in log-vola-equation//
                           // length of theta = (alpha_1,...,alpha_NCOVZ,phi,sigma)//
			   // = NCOVZ + 2                                          //
#define NCOVX 2            // length of covariate vector X (including intercept)   //
#define NCOVZ 2            // length of covariate vector Z (Z never contains an    //
                           // intercept!)                                          //
#define NMATRIX 2          // set to 1 if NCOVZ=0, else set to NCOVZ               //
#define NNCP 6             // number of response categories minus 1                //
#define ITER 4000          // number of iterations to do                           //
#define BURNIN 1000        // number of burnin-steps                               //
#define TAKENORMALERRORS 1 // if 1, normal errors, if 0 t errors are assumed       //
#define USESIMULATION 0    // if real data is to be estimated, type 0, otherwise 1 //
//---------------------------------------------------------------------------------//
//                             PARAMETERS FOR SIMULATION                           //
//---------------------------------------------------------------------------------//
#define SIMUPHI      0.90  //                                                      //
#define SIMUSIGMA    0.20  //                                                      //
                                                                                   //
#define SIMUBETA0    3.50  //                                                      //
#define SIMUBETA1   -0.30  //                                                      //
#define SIMUBETA2   -1.60  //                                                      //
#define SIMUBETA3    0.50  //                                                      //
#define SIMUBETA4    0.00  //                                                      //
#define SIMUBETA5    0.00  //                                                      //
#define SIMUBETA6    0.00  //                                                      //
#define SIMUBETA7    0.00  //                                                      //
#define SIMUBETA8    0.00  //                                                      //
#define SIMUBETA9    0.00  //                                                      //
                                                                                   //
#define SIMUALPHA1   0.25  //                                                      //
#define SIMUALPHA2   0.15  //                                                      //
#define SIMUALPHA3   0.00  //                                                      //
#define SIMUALPHA4   0.00  //                                                      //
#define SIMUALPHA5   0.00  //                                                      //
#define SIMUALPHA6   0.00  //                                                      //
                                                                                   //
#define SIMUCUTP1    0.00  //                                                      //
#define SIMUCUTP2    0.90  //                                                      //
#define SIMUCUTP3    1.80  //                                                      //
#define SIMUCUTP4    2.75  //                                                      //
#define SIMUCUTP5    3.65  //                                                      //
#define SIMUCUTP6    4.50  //                                                      //
                           // covariates which are to be used in simuation must be //
			   // defined in function simu(...)!                       //
//---------------------------------------------------------------------------------//
//                                STARTING VALUES                                  //
//---------------------------------------------------------------------------------//
#define PHISTART     0.85  //                                                      //
#define SIGMASTART   0.30  //                                                      //
#define NUSTART      10    //                                                      //
                                                                                   //
#define ALPHA1START  0.00  //                                                      //
#define ALPHA2START  0.00  //                                                      //
#define ALPHA3START  0.00  //                                                      //
#define ALPHA4START  0.00  //                                                      //
#define ALPHA5START  0.00  //                                                      //
#define ALPHA6START  0.00  //                                                      //
                                                                                   //
#define CP2START     1.00  //                                                      //
#define CP3START     2.00  //                                                      //
#define CP4START     3.00  //                                                      //
#define CP5START     4.00  //                                                      //
#define CP6START     5.00  //                                                      //
//---------------------------------------------------------------------------------//
//                            HYPERPARAMETERS FOR PRIORS                           //
//---------------------------------------------------------------------------------//
#define BETAVARIANCE 10.0000 // each beta_j  has prior N(0,BETAVARIANCE)           //
#define ALPHABOUND   100.000 // each alpha_j has pr. Unif(-ALPHABOUND,ALPHABOUND)  //
#define SIGMABOUND   10.0000 // sigma has prior Unif(0,SIGMABOUND)                 //
//---------------------------------------------------------------------------------//
//                              INPUT AND OUTPUT FILES                             //
//---------------------------------------------------------------------------------//
char procsource[]="yproc.txt";          // input file for response                 //
char desgX[]="xdesg.txt";               // input file for design X (tab-seperated) //
char desgZ[]="zdesg.txt";               // input file for design Z (tab-seperated) //
                                        //                                         //
char paraOUT[]="osvtPARA.dat";	        // output file for parameter estimates     //
char volasOUT[]="osvtVOLAS.dat";        // output file for est. volatilities       //
char logvolasOUT[]="osvtLOGVOLAS.dat";  // output file for est. log-volatilities   //
                                        //                                         //
char procSIMlatent[]="osvtY.dat";       // output file for lat. variables in sim.  //
char procSIM[]="osvtYOBS.dat";          // output file for simulated process       //
//---------------------------------------------------------------------------------//
//---------------------------------------------------------------------------------//
//                                 ADVANCED OPTIONS                                //
//---------------------------------------------------------------------------------//
#define MUFIX -0.60         // model parameter mu must be fixed for identifiab.    //
#define GMSTEPS 1           // if set to 1, program uses GM-steps; if 0, not       //
#define GMSTART 0           // GM-steps will be used only after iteration GMSTART  //
#define NUDRAWSTART 50      // nu will be drawn only after iteration NUDRAWSTART   //
#define PHIDRAWSTART 0      // phi will be drawn only after iteration PHIDRAWSTART //
#define USEPSEUDOAUTOREGR 1 // see function SIMU() !                               //
#define USECHANGEDHTEQUATION 1 // if 1, covariates Z only have impact on present   //
#define DFDF 15             // degr. of freedom of t-distr. (only for simulation)  //
#define PRECISION 0.0000001 // for minimization-algorithm                          //
#define MAXMINSTEPS 200     // for minimization-algorithm                          //
#define GRID 0.001          // grid distance for computation of the Hessian matrix //
#define CONTROLONSCREEN 0   // if 1, current control info is printed on screen     //
#define SCREENUPDATEINTERVAL 1 // used only if CONTROLONSCREEN equals 0            //
//---------------------------------------------------------------------------------//
//---------------------------------------------------------------------------------//
//---------------------------------------------------------------------------------//


//-----------------------------------------------------------------------------------
// Declaration of variables and matrices

Matrix Z(TTTT,NMATRIX);
Matrix ALPHAest(NMATRIX,1);

double minhist[21][ITER+2];
double alphahist[11][ITER+2];
double avg[21];
double avgalpha[11];
double h00=0.00000, p00=0.00000;
double H[TTTT+2];
double LAMBDA[TTTT+2];
double U[TTTT+2];
double Y[TTTT+2];
double Yest[TTTT+2];
unsigned int YOBS[TTTT+2];
double CP[10];
double CPest[10];
Matrix X(TTTT,NCOVX);
Matrix BETA(12,1);
Matrix BETAest(NCOVX,1);
Matrix UNIT(NCOVX,NCOVX);
Matrix XT(NCOVX,1);
Matrix INTER1(NCOVX,NCOVX);
Matrix INTER2(NCOVX,NCOVX);
Matrix INTER3(NCOVX,NCOVX);
Matrix INTER3INV(NCOVX,NCOVX);
Matrix INTER4(NCOVX,1);
Matrix INTER5(NCOVX,1);
Matrix ALPHA(12,1);
double YAST[TTTT+2];
double NORMAL[TTTT+2];
double HESSE[11][11];
unsigned int sest[TTTT+2];
double Hest[TTTT+2];
double LogVolaEST[TTTT+2];
double VolaEST[TTTT+2];
double LAMBDAest[TTTT+2];
unsigned int nuest=NUSTART;
double phiest=0.90000, sigma1est=0.20000, accnu=0.00000;
double mufix=MUFIX;
double htt[TTTT+2], httm1[TTTT+2], kt[TTTT+2], fttm1[TTTT+2], pttm1[TTTT+2];
double ptt[TTTT+2], et[TTTT+2],nt[TTTT+2],rt[TTTT+2],ut[TTTT+2],dt[TTTT+2];
double ct[TTTT+2], zetat[TTTT+2],bt[TTTT+2];

Matrix HESSEM(NNNN,NNNN);
Matrix HESSEINV(NNNN,NNNN);
Matrix MU(NNNN,1);
Matrix MVN(NNNN,1);
Matrix THETA(NNNN,1);

unsigned int ACTUALITER=0;
double PI   = 3.14159265358979;

//-----------------------------------------------------------------------------------
// constants for the seven-component mixture approximation

double qmix[9]     ={ 0.00000,  0.00730, 0.10556, 0.00002, 0.04395, 0.34001, 0.24566, 
                      0.25750,  0.00000};
double qmixacc[9]  ={ 0.00000,  0.00730, 0.11286, 0.11288, 0.15683, 0.49684, 0.74250, 
                      1.00000,  0.00000};
double mumix[9]    ={ 0.00000,-11.40039,-5.24321,-9.83726, 1.50746,-0.65098, 0.52478,
                     -2.35859,  0.00000};
double sigma2mix[9]={ 0.00000,  5.79596, 2.61369, 5.17950, 0.16735, 0.64009, 0.34023, 
                      1.26261,  0.00000};
double sigmamix[9] ={ 0.00000,  2.40748, 1.61669, 2.27585, 0.40908, 0.80006, 0.58329, 
                      1.12366, 0.00000};

//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------

int main(void)
{
   // Declaration of variables
   int i,j,k,l,ncovz;
   ncovz=NCOVZ;
   double frac1=0.00000, frac2=0.00000, acceptprob=0.50000;
   double interc;
   double dgm1=0.00000, dgm2=0.00000, dgm3=0.00000, dgmgamma=0.00000;
   FILE *stream;   
   FILE *procstream;  
   FILE *proc;
   FILE *xdesg;
   FILE *zdesg; 
   unsigned int ui1;
   double f1,f2,f3,f4;
   for (i=1; i<=TTTT; i++) LogVolaEST[i]=0.000000;
   for (i=1; i<=TTTT; i++) VolaEST[i]=0.000000;
   FILE *LogVolas;
   FILE *Volas;
   
   // for fast evaluation of student-t density -> see random.h
   savegammafactors();
   
   // generate unit matrix
   for (i=1; i<=NCOVX; i++) for (k=1; k<=NCOVX; k++) UNIT(i,k)=0.00000;
   for (i=1; i<=NCOVX; i++) UNIT(i,i)=1.00000;      
   
   // open output file and generate head line 
   stream = fopen(paraOUT,"w+");  
   if (TAKENORMALERRORS==1) fprintf(stream,"phi \t sigma \t ");
   else                     fprintf(stream,"phi \t sigma \t nu \t ");
   
   for (i=0; i<NCOVX-1; i++) fprintf(stream,"beta%d \t ",i);
   if (NCOVZ==0 && NNCP<2)   
   {
      if (GMSTEPS==1) 
      { 
         fprintf(stream,"beta%d \t",NCOVX-1);
	 fprintf(stream,"gamma \n");
      }
      else
         fprintf(stream,"beta%d \n",NCOVX-1); 
   }
   else  
   {
      fprintf(stream,"beta%d \t",NCOVX-1);
      if (NCOVZ==0)
      {
         for (i=2; i<NNCP; i++) fprintf(stream,"cutpoint%d \t",i);
	 if (GMSTEPS==1) 
	 { 
	    fprintf(stream,"cutpoint%d \t",NNCP);
	    fprintf(stream,"gamma \n");
	 }
	 else
	    fprintf(stream,"cutpoint%d \n",NNCP);
      }
      if (NNCP<2) 
      {
         for (i=1; i<NCOVZ; i++) fprintf(stream,"alpha%d \t",i);
	 if (GMSTEPS==1) 
	 { 
	    fprintf(stream,"alpha%d \t",NCOVZ);
	    fprintf(stream,"gamma \n");
	 }
	 else
	    fprintf(stream,"alpha%d \n",NCOVZ);
      }
      if (NCOVZ>0 && NNCP>=2)
      {
         for (i=1; i<=NCOVZ; i++) fprintf(stream,"alpha%d \t ",i);
      	 for (i=2; i<NNCP; i++)   fprintf(stream,"cutpoint%d \t",i);
	 if (GMSTEPS==1) 
	 { 
	    fprintf(stream,"cutpoint%d \t",NNCP);
	    fprintf(stream,"gamma \n");
	 }
	 else
	    fprintf(stream,"cutpoint%d \n",NNCP);
      }
   }      
     
   // set random number generator on a random position
   randomize();
   
   if (USESIMULATION==1)
   {
      // simulate process
      simu(MUFIX,SIMUPHI,SIMUSIGMA); 
      // write simulated process in corresponding file 
      procstream = fopen(procSIM,"w+"); 
      for (i=1; i<=TTTT; i++) fprintf(procstream,"%d \n",YOBS[i]);
      fclose(procstream);  
   }
   else
   {
      // read input file which contains the observations
      proc  = fopen(procsource,"r"); 
      for (i=1; i<=TTTT; i++)
      {
         fscanf(proc,"%d\n",&ui1);
	 YOBS[i]=ui1;    
      }
      fclose(proc);
      
      // read input file which contains x-design
      xdesg = fopen(desgX,"r"); 
      for (i=1; i<=TTTT; i++)
      {
         for (k=1; k<NCOVX; k++)
	 {
            fscanf(xdesg,"%lf\t",&f1);
	    if (k==1) X(i,k)=1.000000; 
	    else      X(i,k)=f1;
	 }  
	 fscanf(xdesg,"%lf\n",&f1);
	 X(i,NCOVX)=f1; 
      }
      fclose(xdesg);
      
      // read input file which contains z-design
      zdesg = fopen(desgZ,"r"); 
      for (i=1; i<=TTTT; i++)
      {
         for (k=1; k<NCOVZ; k++)
	 {
            fscanf(zdesg,"%lf\t",&f3);
	    Z(i,k)=f3;
	 }
	 fscanf(zdesg,"%lf\n",&f3);   
	 Z(i,NCOVZ)=f3;   											    
      }        
      fclose(zdesg);   
   }  
   //-----------------------------------------------------  
   // initialize parameters and variables
   
   // initialize LAMBDAest
   if (TAKENORMALERRORS==0) 
      for (i=1; i<=TTTT; i++) 
         LAMBDAest[i]=rand_gamma(0.50000*NUSTART,0.50000*NUSTART);
   else                     
      for (i=1; i<=TTTT; i++) 
         LAMBDAest[i]=1.00000;
   
   // initialize Hest
   Hest[0]=mufix;
   for (i=1; i<=TTTT; i++) 
   {
      interc=0.00000;
      if (NCOVZ>0) interc += ALPHA1START*Z(i,1);
      if (NCOVZ>1) interc += ALPHA2START*Z(i,2);
      if (NCOVZ>2) interc += ALPHA3START*Z(i,3);
      if (NCOVZ>3) interc += ALPHA4START*Z(i,4);
      if (NCOVZ>4) interc += ALPHA5START*Z(i,5);
      if (NCOVZ>5) interc += ALPHA6START*Z(i,6);      
      Hest[i]=mufix+interc+PHISTART*(Hest[i-1]-mufix)+
                           SIGMASTART*normal(0.00000,1.00000);
   }
   
   // initialize CPest
   CPest[1]=0.000000; // this cutpoint is fixed for reasons of identifiability!
   CPest[2]=CP2START; 
   CPest[3]=CP3START;
   CPest[4]=CP4START;
   CPest[5]=CP5START;
   CPest[6]=CP6START;
   
   // initialize Yest
   for (i=1; i<=TTTT; i++)
   {
      switch (YOBS[i])
      {
         case 1: Yest[i] = -0.500000; break;
         case 2: Yest[i] = 0.500000*(CPest[1]+CPest[2]); break;
         case 3: Yest[i] = 0.500000*(CPest[2]+CPest[3]); break;
         case 4: Yest[i] = 0.500000*(CPest[3]+CPest[4]); break;
         case 5: Yest[i] = 0.500000*(CPest[4]+CPest[5]); break;
         case 6: Yest[i] = 0.500000*(CPest[5]+CPest[6]); break;
         case 7: Yest[i] = CPest[6]+0.500000; break;
      }
   }    
   
   //-----------------------------------------------------  
   // Gibbs sampler loop
   
   for (i=1; i<=ITER; i++)
   {
      ACTUALITER=i;
      if (CONTROLONSCREEN==1) printf("\n");
      if (CONTROLONSCREEN==1) printf("iteration: %d \n",i);      
      else 
         if (ACTUALITER%SCREENUPDATEINTERVAL==0) printf("iteration: %d \n",i);
      //---------------------------------------------
      // beta-update
      
      if (CONTROLONSCREEN==1) printf("draw_BETA\n");
      draw_BETA();
      
      //---------------------------------------------
      // y_t^\ast-update
      
      if (CONTROLONSCREEN==1) printf("draw_Y\n");
      draw_Y();      
      
      //---------------------------------------------
      // cupoint-update
      
      if (CONTROLONSCREEN==1) printf("draw_CP\n");
      draw_CP(); 
      
      //---------------------------------------------
      // GM-STEP
      
      dgmgamma = 1.000000;
      if (GMSTEPS==1 && ACTUALITER>GMSTART)
      {
         dgm1=0.50000*(TTTT+NNCP+NCOVX);   // NNCP+1 = K; NCOVX-1 = p, 
	                                   // therefore K+p = NNCP+NCOVX;  now dgm1=a
      
         dgm2=0.00000;
         for (k=1; k<=NCOVX; k++)  
	    dgm2 += BETAest(k,1)*BETAest(k,1)*(1.00000/BETAVARIANCE);
      
         for (k=1; k<=TTTT; k++)
         {
            dgm3=0.00000;
            for (j=1; j<=NCOVX; j++)  dgm3 += BETAest(j,1)*X(k,j); //   dgm3=x_k'beta
            dgm2 += (Yest[k]-dgm3)*(Yest[k]-dgm3)*LAMBDAest[k]/exp(Hest[k]);
         }
      
         dgm2 = 0.50000*dgm2; // now dgm2=b;
      
         dgmgamma=rand_gamma(dgm1,dgm2);
	 dgmgamma=sqrt(dgmgamma);
	 if (dgmgamma>1.050000) dgmgamma=1.050000;
	 if (dgmgamma<0.952381) dgmgamma=0.952381;
	 
	 if (CONTROLONSCREEN==1) printf("GAMMA: %f\n",dgmgamma);
	 
	 for (k=1; k<=NCOVX; k++)  BETAest(k,1)= dgmgamma*BETAest(k,1);
	 for (k=2; k<=NNCP;  k++)  CPest[k]    = dgmgamma*CPest[k];
	 for (k=1; k<=TTTT;  k++)  Yest[k]     = dgmgamma*Yest[k];
	 
      }                
      
      //---------------------------------------------
      // mixture index update
      
      if (CONTROLONSCREEN==1) printf("draw_s\n");
      draw_s();
      
      //---------------------------------------------
      // MH-step for update of phi, sigma, alpha_j, j=1,...,q
      
      if (CONTROLONSCREEN==1) printf("draw_THETA\n");
      // Search the argument, which maximizes log(g(theta)), 
      // that is, which minimizes -log(g(theta))
      if (i==1) 
      {
         switch (ncovz)
	 {
	    case 0: neldermead(NNNN,PHISTART,SIGMASTART); break;
	    case 1: neldermead(NNNN,PHISTART,SIGMASTART,ALPHA1START); break; 
	    case 2: neldermead(NNNN,PHISTART,SIGMASTART,ALPHA1START,ALPHA2START); 
	            break; 
	    case 3: neldermead(NNNN,PHISTART,SIGMASTART,ALPHA1START,ALPHA2START,
	                       ALPHA3START); break; 
	    case 4: neldermead(NNNN,PHISTART,SIGMASTART,ALPHA1START,ALPHA2START,
	                       ALPHA3START,ALPHA4START); break; 
	    case 5: neldermead(NNNN,PHISTART,SIGMASTART,ALPHA1START,ALPHA2START,
	                       ALPHA3START,ALPHA4START,ALPHA5START); break; 
	    case 6: neldermead(NNNN,PHISTART,SIGMASTART,ALPHA1START,ALPHA2START,
	                       ALPHA3START,ALPHA4START,ALPHA5START,ALPHA6START); 
	            break; 
	 }	    
      }
      else  
      {
         switch (ncovz)
	 {
	    case 0: neldermead(NNNN,THETA(1,1),THETA(2,1)); break;
	    case 1: neldermead(NNNN,THETA(1,1),THETA(2,1),THETA(3,1)); break;
	    case 2: neldermead(NNNN,THETA(1,1),THETA(2,1),THETA(3,1),THETA(4,1)); 
	            break;
	    case 3: neldermead(NNNN,THETA(1,1),THETA(2,1),THETA(3,1),THETA(4,1),
	                       THETA(5,1)); break;
	    case 4: neldermead(NNNN,THETA(1,1),THETA(2,1),THETA(3,1),THETA(4,1),
	                       THETA(5,1),THETA(6,1)); break;
	    case 5: neldermead(NNNN,THETA(1,1),THETA(2,1),THETA(3,1),THETA(4,1),
	                       THETA(5,1),THETA(6,1),THETA(7,1)); break;
	    case 6: neldermead(NNNN,THETA(1,1),THETA(2,1),THETA(3,1),THETA(4,1),
	                       THETA(5,1),THETA(6,1),THETA(7,1),THETA(8,1)); break;	    
	 }
      }
               
      for (k=1; k<=2; k++)      minhist[k][i]  = min[k];
      for (k=1; k<=NCOVZ; k++)  alphahist[k][i]= min[2+k];
      
      // Compute the Hessian at the minimum
      switch (ncovz)
      {
         case 0: Hesse(NNNN,GRID,min[1],min[2]); break;
	 case 1: Hesse(NNNN,GRID,min[1],min[2],min[3]); break;
	 case 2: Hesse(NNNN,GRID,min[1],min[2],min[3],min[4]); break;
	 case 3: Hesse(NNNN,GRID,min[1],min[2],min[3],min[4],min[5]); break;
	 case 4: Hesse(NNNN,GRID,min[1],min[2],min[3],min[4],min[5],min[6]); break;
	 case 5: Hesse(NNNN,GRID,min[1],min[2],min[3],min[4],min[5],min[6],min[7]); 
	         break;
	 case 6: Hesse(NNNN,GRID,min[1],min[2],min[3],min[4],min[5],min[6],min[7],
	               min[8]); break;
      }
      
      // Copy the arrays into matrices
      for (k=1; k<=NNNN; k++) MU(k,1)=min[k];
      for (k=1; k<=NNNN; k++) for (l=1; l<=NNNN; l++) HESSEM(k,l)=HESSE[k][l];
      if (CONTROLONSCREEN==1) 
         printf("Hessian matrix pos.def.: %d \n",HESSEM.posdef());
      
      // Draw a proposal MVN from the multivariate normal distribution: 
      HESSEINV=HESSEM.invertPosDef();
      MVN=mvnormal(MU,HESSEINV);
      if (CONTROLONSCREEN==1) 
      {
         printf("min phi  : %f \n",MU(1,1));
         printf("min sigma: %f \n",MU(2,1));      
         if (NCOVZ>0) printf("min alpha1: %f \n",MU(3,1));
         if (NCOVZ>1) printf("min alpha2: %f \n",MU(4,1));
         if (NCOVZ>2) printf("min alpha3: %f \n",MU(5,1));
         if (NCOVZ>3) printf("min alpha4: %f \n",MU(6,1));
         if (NCOVZ>4) printf("min alpha5: %f \n",MU(7,1));
         if (NCOVZ>5) printf("min alpha6: %f \n",MU(8,1));
      }
     
      // We need theta_0 for computing the acceptance probability
      if (i==1)  for (k=1; k<=NNNN; k++) THETA(k,1)=min[k];     
      
      // Compute the acceptance probability
      switch (ncovz)
      {
         case 0: frac1 = exp(-func(MVN(1,1),MVN(2,1))
	                     +func(THETA(1,1),THETA(2,1))); break;
	 case 1: frac1 = exp(-func(MVN(1,1),MVN(2,1),MVN(3,1))
	                     +func(THETA(1,1),THETA(2,1),THETA(3,1))); break;
	 case 2: frac1 = exp(-func(MVN(1,1),MVN(2,1),MVN(3,1),MVN(4,1))
	                     +func(THETA(1,1),THETA(2,1),THETA(3,1),THETA(4,1))); 
	         break;
	 case 3: frac1 = exp(-func(MVN(1,1),MVN(2,1),MVN(3,1),MVN(4,1),MVN(5,1))
	                     +func(THETA(1,1),THETA(2,1),THETA(3,1),THETA(4,1),
			           THETA(5,1))); 
		 break;
	 case 4: frac1 = exp(-func(MVN(1,1),MVN(2,1),MVN(3,1),MVN(4,1),MVN(5,1),
	                           MVN(6,1))
	                     +func(THETA(1,1),THETA(2,1),THETA(3,1),THETA(4,1),
			           THETA(5,1),THETA(6,1))); 
		 break;
	 case 5: frac1 = exp(-func(MVN(1,1),MVN(2,1),MVN(3,1),MVN(4,1),MVN(5,1),
	                           MVN(6,1),MVN(7,1))
	                     +func(THETA(1,1),THETA(2,1),THETA(3,1),THETA(4,1),
			           THETA(5,1),THETA(6,1),THETA(7,1))); 
		 break;
	 case 6: frac1 = exp(-func(MVN(1,1),MVN(2,1),MVN(3,1),MVN(4,1),MVN(5,1),
	                           MVN(6,1),MVN(7,1),MVN(8,1))
	                     +func(THETA(1,1),THETA(2,1),THETA(3,1),THETA(4,1),
			           THETA(5,1),THETA(6,1),THETA(7,1),THETA(8,1))); 
	         break;     
      }
      frac2 = undmvnormal(THETA,MU,HESSEINV)/undmvnormal(MVN,MU,HESSEINV);
      
      acceptprob=minimum(1.00000,frac1*frac2);
      if (CONTROLONSCREEN==1) printf("ACCEPT-PROBABILITY: %f \n",acceptprob);
      minhist[4][i]=acceptprob;      
					      
      // if we accept the proposal MVN, theta changes
      if (uniform(0,1)<=acceptprob)  for (k=1; k<=NNNN; k++) THETA(k,1)=MVN(k,1);
      
      if (ACTUALITER<=PHIDRAWSTART) THETA(1,1)=PHISTART;
     
      //---------------------------------------------
      // update of log-volatilities
      
      if (CONTROLONSCREEN==1) printf("draw_H\n");
      draw_H(); 
      
      // estimate the log-volatilities and write output-file at iteration ITER
      if (i>BURNIN)
      {
         for (k=1; k<=TTTT; k++) LogVolaEST[k] += Hest[k];
      }
      if (i==ITER)
      {
         for (k=1; k<=TTTT; k++) LogVolaEST[k]=LogVolaEST[k]/(1.00000*(ITER-BURNIN));
	 LogVolas = fopen(logvolasOUT,"w+");
	 fprintf(LogVolas,"logVolatilities \n");
	 for (k=1; k<=TTTT; k++) fprintf(LogVolas,"%f \n",LogVolaEST[k]);
	 fclose(LogVolas);
      }   
      
      // estimate the volatilities and write output-file at iteration ITER
      if (i>BURNIN)
      {
         for (k=1; k<=TTTT; k++) VolaEST[k] += exp(Hest[k]+0.60000);
      }
      if (i==ITER)
      {
         for (k=1; k<=TTTT; k++) VolaEST[k] = VolaEST[k]/(1.00000*(ITER-BURNIN));
	 Volas = fopen(volasOUT,"w+");
	 fprintf(Volas,"NormalizedVolatilities \n");
	 for (k=1; k<=TTTT; k++) fprintf(Volas,"%f \n",VolaEST[k]);
	 fclose(Volas);
      }      
      
      //---------------------------------------------
      // nu-update (only for OSVt-model)
      
      if (ACTUALITER<=NUDRAWSTART || TAKENORMALERRORS==1) nuest=NUSTART;
      else
      {
         if (CONTROLONSCREEN==1) printf("draw_nu\n");
         draw_nu();
      }
      minhist[5][ACTUALITER]=(double) nuest;
      minhist[6][ACTUALITER]=accnu;   
                  
      //---------------------------------------------
      // lamda-update  (only for OSVt-model)
      
      if (TAKENORMALERRORS==0)
      {
         if (CONTROLONSCREEN==1) printf("draw_LAMBDA\n");
         draw_LAMBDA();
      }
      else  
      {
         for (k=1; k<=TTTT; k++) LAMBDAest[k]=1.00000;
      }
      
      //---------------------------------------------
      // write output-file which contains the parameter estimates
      
      if (TAKENORMALERRORS==1) 
         fprintf(stream,"%f \t %f \t ",THETA(1,1),THETA(2,1));
      else 
         fprintf(stream,"%f \t %f \t %d \t ",THETA(1,1),THETA(2,1),nuest);
      for (k=1; k<NCOVX; k++) fprintf(stream,"%f \t ",BETAest(k,1));
      if (NCOVZ==0 && NNCP<2)   
      {
         if (GMSTEPS==1) 
         { 
            fprintf(stream,"%f \t", BETAest(NCOVX,1));
	    fprintf(stream,"%f \n", dgmgamma);
         }
         else
            fprintf(stream,"%f \n", BETAest(NCOVX,1));
      }
      else  
      {
         fprintf(stream,"%f \t",BETAest(NCOVX,1));
         if (NCOVZ==0)
         {
            for (k=2; k<NNCP; k++) fprintf(stream,"%f \t",CPest[k]);
	    if (GMSTEPS==1) 
	    { 
	       fprintf(stream,"%f \t",CPest[NNCP]);
	       fprintf(stream,"%f \n",dgmgamma);
	    }
	    else
	       fprintf(stream,"%f \n",CPest[NNCP]);
         }
         if (NNCP<2) 
         { 
            for (k=1; k<NCOVZ; k++) fprintf(stream,"%f \t ",THETA(2+k,1));
	    if (GMSTEPS==1) 
	    { 
	       fprintf(stream,"%f \t",THETA(2+NCOVZ,1)); 
	       fprintf(stream,"%f \n",dgmgamma);
	    }
	    else
	       fprintf(stream,"%f \n",THETA(2+NCOVZ,1)); 
         }
         if (NCOVZ>0 && NNCP>=2)
         {
            for (k=1; k<=NCOVZ; k++) fprintf(stream,"%f \t ",THETA(2+k,1));
      	    for (k=2; k<NNCP; k++)   fprintf(stream,"%f \t",CPest[k]);
	    if (GMSTEPS==1) 
	    { 
	       fprintf(stream,"%f \t",CPest[NNCP]);
	       fprintf(stream,"%f \n",dgmgamma);
	    }
	    else
	       fprintf(stream,"%f \n",CPest[NNCP]);
         }
      }                                 
   }
   
   // write results on screen
   for (i=BURNIN+1; i<=ITER; i++) 
      for (k=1; k<=6+NCOVX; k++)    avg[k]      += minhist[k][i];
   for (i=BURNIN+1; i<=ITER; i++) 
      for (k=1; k<=NCOVZ; k++)      avgalpha[k] += alphahist[k][i];
   printf("\n\nPosterior mean estimates:\n"); 
   printf("phi   :\t%f\n",avg[1]/(ITER-BURNIN)); 
   printf("sigma :\t%f\n",avg[2]/(ITER-BURNIN)); 
   if (TAKENORMALERRORS==0) printf("nu    :\t%f\n",avg[5]/(ITER-BURNIN));
   for (i=1; i<=NCOVX; i++) printf("beta%d :\t%f\n",i-1,avg[6+i]/(ITER-BURNIN)); 
   for (i=1; i<=NCOVZ; i++) printf("alpha%d:\t%f\n",i,avgalpha[i]/(ITER-BURNIN)); 
   printf("\nAverage of acceptance probabilities:\n"); 
   printf("theta:\t%f\n",avg[4]/(ITER-BURNIN));   
   if (TAKENORMALERRORS==0) printf("nu   :\t%f\n",avg[6]/(ITER-BURNIN)); 
   printf("\n");
   
   fclose(stream);
}

//-----------------------------------------------------------------------------------

void Hesse(unsigned int n, double griddist, 
           double x1, double x2, double x3, double x4, double x5,  
           double x6, double x7, double x8, double x9, double x10)
{   
   int i,j,k;
   double x[11], xl[11], xr[11], xlo[11], xro[11], xlu[11], xru[11];  
   double xx, xxr, xxl, xxro, xxlo, xxru, xxlu; 
   x[1]=x1; x[2]=x2; x[3]=x3; x[4]=x4; x[5]=x5; 
   x[6]=x6; x[7]=x7; x[8]=x8; x[9]=x9; x[10]=x10;
   
   for (i=1; i<=n; i++) for (k=1; k<=n; k++) HESSE[i][k]=0.00000;
   
   xx = func(x1,x2,x3,x4,x5,x6,x7,x8,x9,x10); 
      
   // Compute diagonal of matrix
   for (i=1; i<=n; i++)
   {
      for (k=1; k<=n; k++) xl[k] = x[k];
      for (k=1; k<=n; k++) xr[k] = x[k];
      xl[i] = x[i] - griddist;
      xr[i] = x[i] + griddist;
      xxl   = func(xl[1],xl[2],xl[3],xl[4],xl[5],xl[6],xl[7],xl[8],xl[9],xl[10]);
      xxr   = func(xr[1],xr[2],xr[3],xr[4],xr[5],xr[6],xr[7],xr[8],xr[9],xr[10]);
      HESSE[i][i] = (xxr-2*xx+xxl)/(griddist*griddist);
   }
   for (i=1; i<=n; i++)
   {
      for (j=1; j<=n; j++)
      {
         if (j!=i)
	 {
            for (k=1; k<=n; k++) xlo[k] = x[k];
	    for (k=1; k<=n; k++) xro[k] = x[k];
	    for (k=1; k<=n; k++) xlu[k] = x[k];
	    for (k=1; k<=n; k++) xru[k] = x[k];
	    xlo[i] = xlo[i]-griddist;
	    xlo[j] = xlo[j]+griddist;
	    xro[i] = xro[i]+griddist;
	    xro[j] = xro[j]+griddist;
	    xlu[i] = xlu[i]-griddist;
	    xlu[j] = xlu[j]-griddist;
	    xru[i] = xru[i]+griddist;
	    xru[j] = xru[j]-griddist;
	    xxro   = func(xro[1], xro[2], xro[3], xro[4], xro[5], 
	                  xro[6], xro[7], xro[8], xro[9], xro[10]);
	    xxlo   = func(xlo[1], xlo[2], xlo[3], xlo[4], xlo[5], 
	                  xlo[6], xlo[7], xlo[8], xlo[9], xlo[10]);
	    xxru   = func(xru[1], xru[2], xru[3], xru[4], xru[5], 
	                  xru[6], xru[7], xru[8], xru[9], xru[10]);
	    xxlu   = func(xlu[1], xlu[2], xlu[3], xlu[4], xlu[5], 
	                  xlu[6], xlu[7], xlu[8], xlu[9], xlu[10]);
	    HESSE[i][j] = (xxro-xxlo-xxru+xxlu)/(4*griddist*griddist);
	    HESSE[j][i] = HESSE[i][j];
	 }
      }
   }
}

//-----------------------------------------------------------------------------------

double func(double x1, double x2, double x3, double x4, double x5, 
            double x6, double x7, double x8, double x9, double x10) // MinusLogG
{
   int k,t;
   double result  = 0.000000;
   double intert  = 0.000000;
   double intertm1= 0.000000;
   int alphajoutofbounds=0;
   
   double muF     = mufix;
   double phiF    = x1;
   double sigma1F = x2;
   
   double alpha[10];
   if (NCOVZ>0) alpha[1] = x3;
   if (NCOVZ>1) alpha[2] = x4;
   if (NCOVZ>2) alpha[3] = x5;
   if (NCOVZ>3) alpha[4] = x6;
   if (NCOVZ>4) alpha[5] = x7;
   if (NCOVZ>5) alpha[6] = x8; 
   
   for (k=1; k<=NCOVZ; k++)
      if (alpha[k]<-ALPHABOUND || alpha[k]>ALPHABOUND) alphajoutofbounds=1;
        
   double sigma2F = sigma1F*sigma1F;
   double phi2F   = phiF*phiF;   
   
   if (sigma1F<0.001 || sigma1F>SIGMABOUND || 
       phiF>=0.999 || phiF<=-0.999 || alphajoutofbounds==1) 
      return (1000000000);
   else
   {
      htt[0]= muF;
      ptt[0]= sigma2F/(1.00000-phi2F);

      for (t=1; t<=TTTT; t++)
      { 
         intert   = 0.00000;
	 for (k=1; k<=NCOVZ; k++)  intert   += alpha[k]*Z(t,k);
	 intertm1 = 0.00000;
	 if (USECHANGEDHTEQUATION==1)
	 {
	    if (t>1) for (k=1; k<=NCOVZ; k++)  intertm1 += alpha[k]*Z(t-1,k);
	 }
	 
         httm1[t] = muF + intert + phiF*(htt[t-1]-muF-intertm1);
         pttm1[t] = phi2F*ptt[t-1] + sigma2F;  
	  
         fttm1[t] = pttm1[t] + sigma2mix[sest[t]];
         kt[t]    = pttm1[t]/fttm1[t];
         htt[t]   = httm1[t] + kt[t]*(YAST[t]-mumix[sest[t]]-httm1[t]);
         ptt[t]   = (1-kt[t])*pttm1[t];
        
	 result   = result - log(fttm1[t]) - (YAST[t]-mumix[sest[t]]-httm1[t])*
		                          (YAST[t]-mumix[sest[t]]-httm1[t])/fttm1[t];
      }
   
      return (-0.50000*result);
   } 
} 

//-----------------------------------------------------------------------------------

void simu(double mu, double phi, double sigma1)
{  
   int t,i,k;
   double x0       = 0.00000;
   double z0       = 0.00000;
   double inter    = 0.00000;
   double interz   = 0.00000;
   double interzm1 = 0.00000;
   FILE *streamY; 
   streamY = fopen(procSIMlatent,"w+");  
   
   
   FILE *xde;
   FILE *zde; 
   double f11,f12,f13,f14;
   
   // read given (simulated or real) files with covariates
   xde = fopen(desgX,"r"); 
   for (i=1; i<=TTTT; i++)
   {
      for (k=1; k<NCOVX; k++)
      {
         fscanf(xde,"%lf\t",&f11);
         if (k==1) X(i,k)=1.000000;
	 else      X(i,k)=f11;
      }
      fscanf(xde,"%lf\n",&f11);
      X(i,NCOVX)=f11;  
   }
   fclose(xde);
   
   zde = fopen(desgZ,"r"); 
   for (i=1; i<=TTTT; i++)
   {
      for (k=1; k<NCOVZ; k++)
      {
         fscanf(zde,"%lf\t",&f13);
         Z(i,k)=f13;
      }
      fscanf(zde,"%lf\n",&f13);
      Z(i,NCOVZ)=f13;									    
   }     
 
   fclose(zde);    
   
   // copy the true beta's in vector BETA
   BETA(1,1)=SIMUBETA0; BETA(2,1)=SIMUBETA1; BETA(3,1)=SIMUBETA2; 
   BETA(4,1)=SIMUBETA3; BETA(5,1)=SIMUBETA4; BETA(6,1)=SIMUBETA5; 
   BETA(7,1)=SIMUBETA6; BETA(8,1)=SIMUBETA7; BETA(9,1)=SIMUBETA8; 
   BETA(10,1)=SIMUBETA9;  
   
   // copy the true alpha's in vector ALPHA 
   ALPHA(1,1)=SIMUALPHA1; ALPHA(2,1)=SIMUALPHA2; ALPHA(3,1)=SIMUALPHA3; 
   ALPHA(4,1)=SIMUALPHA4; ALPHA(5,1)=SIMUALPHA5; ALPHA(6,1)=SIMUALPHA6; 
   
   // copy the true cutpoints in vector ALPHA 
   CP[1]=SIMUCUTP1; CP[2]=SIMUCUTP2; CP[3]=SIMUCUTP3; 
   CP[4]=SIMUCUTP4; CP[5]=SIMUCUTP5; CP[6]=SIMUCUTP6;
   
   // generate the needed random variables
   for (t=1; t<=TTTT; t++)  NORMAL[t]=normal(0.00000,sigma1);
   for (t=1; t<=TTTT; t++)  U[t]     =normal(0,1);
   for (t=1; t<=TTTT; t++)  LAMBDA[t]=rand_gamma(DFDF/2,DFDF/2);
   if (TAKENORMALERRORS==1) for (t=1; t<=TTTT; t++)  LAMBDA[t]=1.00000;
   
   // generate process
   H[0]=mu;   
   for (t=1; t<=TTTT; t++)
   {
      // compute the covariates' impact at time t
      inter=0.00000;
      if (USEPSEUDOAUTOREGR==1) 
      {
         if (t==1) X(t,2)=4.00000;
	 else      X(t,2)=(double) YOBS[t-1];      
      }
      for (i=1; i<=NCOVX; i++)           inter    += BETA(i,1) *X(t,i);
      interz=0.00000;
      for (i=1; i<=NCOVZ; i++)           interz   += ALPHA(i,1)*Z(t,i);
      interzm1 = 0.00000;
      if (USECHANGEDHTEQUATION==1)
      {
         if (t>1) for (i=1; i<=NCOVZ; i++)  interzm1 += ALPHA(i,1)*Z(t-1,i);
      }
      
      // compute H[t] and Y[t] at time t      
      H[t] = mu + interz + phi*(H[t-1]-mu-interzm1) + NORMAL[t];   
      Y[t] = inter + exp(0.5000*H[t]) * U[t] * sqrt(LAMBDA[t]); 
      
      if (Y[t]<CP[1]) YOBS[t]=1;
      else
         if (Y[t]<CP[2] || NNCP==1) YOBS[t]=2;
	 else 
	    if (Y[t]<CP[3] || NNCP==2) YOBS[t]=3;
	    else
	       if (Y[t]<CP[4] || NNCP==3) YOBS[t]=4;
	       else
	          if (Y[t]<CP[5] || NNCP==4) YOBS[t]=5;
		  else
		     if (Y[t]<CP[6] || NNCP==5) YOBS[t]=6;
		     else YOBS[t]=7;      
      
      if (t<10 && CONTROLONSCREEN==1) 
         printf("H[%d]: %f \t impact of alpha: %f \n",t,H[t],interz);
      fprintf(streamY,"%f \t %d \n",Y[t],YOBS[t]);
   }
   
   fclose(streamY);
}  

//-----------------------------------------------------------------------------------

double impactX(unsigned int t)
{   
   int i;
   double inter=0.000000;
   if (NCOVX==0) return 0.000000;
   else 
   {
      for (i=1; i<=NCOVX; i++) inter+=BETAest(i,1)*X(t,i);
      return inter;
   }
}   

//-----------------------------------------------------------------------------------

void draw_CP(void)
{
   int i,t;
   double left=0.000000, right=0.000000;
   for (i=2; i<=NNCP; i++)
   {
      left = CPest[i-1];
      if (i<NNCP) right = CPest[i+1]; else right=1000.0;
      for (t=1; t<=TTTT; t++)
      {
         if (YOBS[t]==i)   if (Yest[t]>left)  left  = Yest[t];
	 if (YOBS[t]==i+1) if (Yest[t]<right) right = Yest[t];     
      }  
      CPest[i]=uniform(left,right);
      if (CONTROLONSCREEN==1) 
         printf("CPest[%d] = %f \t left = %f \t right = %f \n",
	        i,CPest[i],left,right);
   }
}

//-----------------------------------------------------------------------------------

void draw_Y(void)
{
   unsigned int k,t;
   double interb=0.000000;
   for (t=1; t<=TTTT; t++)
   {
      interb=0.000000;
      for (k=1; k<=NCOVX; k++) interb += BETAest(k,1)*X(t,k);
            
      switch (YOBS[t])
      {
         case 1: Yest[t] = rs_trunc_normal(         0.00000, interb,
	                                   sqrt(exp(Hest[t])/LAMBDAest[t])); break;
	 case 2: Yest[t] = ds_trunc_normal(0.00000, CPest[2],interb,
	                                   sqrt(exp(Hest[t])/LAMBDAest[t])); break;
	 case 3: Yest[t] = ds_trunc_normal(CPest[2],CPest[3],interb,
	                                   sqrt(exp(Hest[t])/LAMBDAest[t])); break;
	 case 4: Yest[t] = ds_trunc_normal(CPest[3],CPest[4],interb,
	                                   sqrt(exp(Hest[t])/LAMBDAest[t])); break;
	 case 5: Yest[t] = ds_trunc_normal(CPest[4],CPest[5],interb,
	                                   sqrt(exp(Hest[t])/LAMBDAest[t])); break;
	 case 6: Yest[t] = ds_trunc_normal(CPest[5],CPest[6],interb,
	                                   sqrt(exp(Hest[t])/LAMBDAest[t])); break;
	 case 7: Yest[t] = ls_trunc_normal(CPest[6],         interb,
	                                   sqrt(exp(Hest[t])/LAMBDAest[t])); break;
      }         
   }   
}
  
//-----------------------------------------------------------------------------------

void draw_BETA(void)
{   
   unsigned int i,k,t;
   double d1=0.00000,d2=0.00000;
   
   for (i=1; i<=NCOVX; i++) for (k=1; k<=NCOVX; k++) INTER1(i,k)=0.00000; 
   for (i=1; i<=NCOVX; i++) for (k=1; k<=NCOVX; k++) INTER2(i,k)=0.00000; 
   for (i=1; i<=NCOVX; i++) for (k=1; k<=NCOVX; k++) INTER3(i,k)=0.00000; 
   for (i=1; i<=NCOVX; i++) INTER4(i,1)=0.00000; 
   
   for (t=1; t<=TTTT; t++)
   {
      XT     = (X.tmat2(t,1,t,NCOVX)).transp();
      INTER2 = XT*XT.transp();
      d1     = LAMBDAest[t]/exp(Hest[t]);
      INTER3 = INTER3 + d1*INTER2;  
      
      d2     = Yest[t]*d1;
      INTER4 = INTER4 + d2*XT;       
   }
   INTER3 = INTER3 + (1.00000/BETAVARIANCE)*UNIT;
   // INTER 3 is just the matrix B now
   
   INTER3INV = INTER3.invertPosDef();   
   INTER5    = INTER3INV*INTER4;
   
   BETAest   = mvnormal(INTER5,INTER3INV); 
   
   for (i=1; i<=NCOVX; i++) minhist[6+i][ACTUALITER]=BETAest(i,1);      
}   
  
//-----------------------------------------------------------------------------------

void draw_nu(void)
{   
   unsigned int nnu,nustep,k,t;
   double dens[130];
   unsigned int expo[130];
   unsigned int maxfound=0;
   unsigned int proposal=150;
   double aa,bb,vv;
   double ffrac1=0.000000;
   double ffrac2=0.000000;
   int ediff=0;
        
   double result=1.00000;
   double resultold=1.00000;
   double resultnew=1.00000;
   
   
     
   // find approximate argmax for target density  
   nnu=64;
   nustep=32;
   
   result=1.000000;
   dens[nnu]=0.000000;
   expo[nnu]=0;
      
   for (t=1; t<=TTTT; t++)
   {
      result*=dstudent(Yest[t],nnu,impactX(t),exp(0.500000*Hest[t]));
      // avoiding too small or too big results          
      if (result<0.0100000) { result*=100.00000; expo[nnu]+=2; }
      if (result>100.00000) { result/=100.00000; expo[nnu]-=2; }	    
   }
   dens[nnu]=result;
   
   while (maxfound==0)
   {
      result=1.000000;
      dens[nnu-nustep]=0.000000;
      expo[nnu-nustep]=0;
      
      for (t=1; t<=TTTT; t++)
      {
         result*=dstudent(Yest[t],nnu-nustep,impactX(t),exp(0.500000*Hest[t]));
         // avoiding too small or too big results          
         if (result<0.0100000) { result*=100.00000; expo[nnu-nustep]+=2; }
	 if (result>100.00000) { result/=100.00000; expo[nnu-nustep]-=2; }	    
      }
      dens[nnu-nustep]=result;      
      
      result=1.000000;
      dens[nnu+nustep]=0.000000;
      expo[nnu+nustep]=0;      
      
      for (t=1; t<=TTTT; t++)
      {
         result*=dstudent(Yest[t],nnu+nustep,impactX(t),exp(0.500000*Hest[t]));
         // avoiding too small or too big results          
         if (result<0.0100000) { result*=100.00000; expo[nnu+nustep]+=2; }
	 if (result>100.00000) { result/=100.00000; expo[nnu+nustep]-=2; }	    
      }
      dens[nnu+nustep]=result;    
	           
      if (expo[nnu+nustep]<expo[nnu] || 
         (expo[nnu+nustep]==expo[nnu] && dens[nnu+nustep]>dens[nnu])) nnu+=nustep; 
      else
         if (expo[nnu-nustep]<expo[nnu] || 
	    (expo[nnu-nustep]==expo[nnu] && dens[nnu-nustep]>dens[nnu])) nnu-=nustep; 
      
      if (nustep==1) maxfound=nnu;
      else nustep/=2;
   }
   if (CONTROLONSCREEN==1) printf("maxfound=%d\n",maxfound); 
   
   // use discretized gamma distr. with truncation to 0.5 - 127.5 as proposal density
      
   while (proposal>127 || proposal<1)
   {
      vv=0.2*maxfound*maxfound;
      if (maxfound<30) vv=0.10*maxfound*maxfound;
      if (maxfound<10) vv=0.05*maxfound*maxfound;
      bb=maxfound/(2*vv)+sqrt(maxfound*maxfound+4*vv)/2/vv;
      aa=bb*bb*vv;
      // mode of this gamma distribution is now maxfound
      proposal = (int) floor(rand_gamma(aa,bb)+0.50000);   
   }
      
   // accept or not?
   // compute posterior(proposal)/posterior(nuest)
   
   expo[nuest]=0;
   resultold=1.000000;
   
   for (t=1; t<=TTTT; t++)
   {
      resultold*=dstudent(Yest[t],nuest,impactX(t),exp(0.500000*Hest[t]));
      // avoiding too small or too big results          
      if (resultold<0.0100000) { resultold*=100.00000; expo[nuest]+=2; }
      if (resultold>100.00000) { resultold/=100.00000; expo[nuest]-=2; }	    
   }
   
   expo[proposal]=0;
   resultnew=1.000000;
   
   for (t=1; t<=TTTT; t++)
   {
      resultnew*=dstudent(Yest[t],proposal,impactX(t),exp(0.500000*Hest[t]));
      // avoiding too small or too big results          
      if (resultnew<0.0100000) { resultnew*=100.00000; expo[proposal]+=2; }
      if (resultnew>100.00000) { resultnew/=100.00000; expo[proposal]-=2; }	    
   }   
   
   ediff=expo[nuest]-expo[proposal];
   ffrac1=resultnew/resultold;
   if (ediff<0) for (k=0; k<-ediff; k+=2) ffrac1*=0.010000;
   if (ediff>0) for (k=0; k<ediff;  k+=2) ffrac1*=100.0000; 
   
   // compute proposal-density(nuest)/proposal-density(proposal)
   // via Newton-Cotes-formula
   ffrac2=unddisgamma(nuest,aa,bb)/unddisgamma(proposal,aa,bb); 
   // via linear approximation
   // ffrac2=pow(1.00000*nuest/(1.00000*proposal),aa-1)*
   //        exp(bb*(-1.00000*nuest+1.00000*proposal));
   
   // compute acceptance probability
   accnu=minimum(1.00000,ffrac1*ffrac2);
   
   if (CONTROLONSCREEN==1) 
   {
      printf("resultold=%f; expo=%d; nuest   =%d\n",resultold,expo[nuest],nuest); 
      printf("resultnew=%f; expo=%d; proposal=%d\n",
             resultnew,expo[proposal],proposal); 
      printf("aa=%f\n",aa); 
      printf("bb=%f\n",bb);    
      printf("ffrac1=%f\n",ffrac1); 
      printf("ffrac2=%f\n",ffrac2);
   }    
   
   if (uniform(0,1)<=accnu) nuest=proposal;
   else                     nuest=nuest;   
   if (CONTROLONSCREEN==1) 
      printf("accept proposal with prob=%f; nuest=%d\n",accnu,nuest); 
   
   if (ACTUALITER<=NUDRAWSTART) nuest=NUSTART;
}   


//-----------------------------------------------------------------------------------

void draw_LAMBDA(void)
{   
   unsigned int t;
   
   for (t=1; t<=TTTT; t++)  
      LAMBDAest[t]=rand_gamma(0.50000*(nuest+1.00000),
                              0.50000*(nuest+(Yest[t]-impactX(t))*
			                     (Yest[t]-impactX(t))/exp(Hest[t])));
}   

//-----------------------------------------------------------------------------------

void draw_H(void)
{  
   int k,t;
   double intert  = 0.000000;
   double intertm1= 0.000000;
   
   double muF     = mufix;
   double phiF    = THETA(1,1);
   double sigma1F = THETA(2,1);
   
   unsigned int counter=0;
   
   double alpha[10];
   if (NCOVZ>0) alpha[1] = THETA(3,1);
   if (NCOVZ>1) alpha[2] = THETA(4,1);
   if (NCOVZ>2) alpha[3] = THETA(5,1);
   if (NCOVZ>3) alpha[4] = THETA(6,1);
   if (NCOVZ>4) alpha[5] = THETA(7,1);
   if (NCOVZ>5) alpha[6] = THETA(8,1);      
     
   double sigma2F = sigma1F*sigma1F;
   double phi2F   = phiF*phiF;   
   
   htt[0]= muF;
   ptt[0]= sigma2F/(1.00000-phi2F);

   for (t=1; t<=TTTT; t++)
   { 
      intert   = 0.00000;
      for (k=1; k<=NCOVZ; k++)  intert += alpha[k]*Z(t,k);
      intertm1 = 0.00000;
      if (USECHANGEDHTEQUATION==1)
      {
         if (t>1) for (k=1; k<=NCOVZ; k++)  intertm1 += alpha[k]*Z(t-1,k);
      }
      
      httm1[t] = muF + intert + phiF*(htt[t-1]-muF-intertm1);
      pttm1[t] = phi2F*ptt[t-1] + sigma2F;   
      
      fttm1[t] = pttm1[t]+sigma2mix[sest[t]];
      kt[t]    = pttm1[t]/fttm1[t];
      htt[t]   = httm1[t]+kt[t]*(YAST[t]-mumix[sest[t]]-httm1[t]);
      ptt[t]   = (1-kt[t])*pttm1[t];
        
      et[t]    = YAST[t]-mumix[sest[t]]-httm1[t];
   }
   // now all e_t's, fttm1_t's, and k_t's are known  
   
   rt[TTTT]=0.00000;
   ut[TTTT]=0.00000;
   
   counter=0;
   
   for (t=TTTT; t>=1; t--)
   { 
      nt[t]    = 1.00000/fttm1[t] + phiF*phiF*kt[t]*kt[t]*ut[t];
      dt[t]    = et[t]/fttm1[t] - rt[t]*phiF*kt[t];
      
      ct[t]    = sigma2mix[sest[t]] - sigma2mix[sest[t]]*sigma2mix[sest[t]]*nt[t];
      
      if (ct[t]>=0.00000) zetat[t] = normal(0.00000,sqrt(ct[t]));
      else                
      {
         zetat[t] = 0.00000;
         counter++;	 
      }
                  
      bt[t]    = sigma2mix[sest[t]]*( nt[t] - phiF*phiF*kt[t]*ut[t]);
      
      rt[t-1]  = et[t]/fttm1[t]   + (phiF - phiF*kt[t])*rt[t] - bt[t]*zetat[t]/ct[t];
      ut[t-1]  = 1.00000/fttm1[t] + (phiF - phiF*kt[t])*(phiF - phiF*kt[t])*ut[t] + 
                 bt[t]*bt[t]/ct[t];
           
      Hest[t]  = YAST[t] - mumix[sest[t]] - sigma2mix[sest[t]]*dt[t] - zetat[t];
      
      if (Hest[t]<-20) 
      { 
         Hest[t]=-20;
	 if (CONTROLONSCREEN==1) printf("H[%d] set to -20\n",t);
      }
      if (Hest[t]>  5) 
      {
         Hest[t]=5;
	 if (CONTROLONSCREEN==1) printf("H[%d] set to 5\n",t);
      }
   }  
   if (CONTROLONSCREEN==1) printf("zeta[t]-corrections: %d \n",counter); 
}   

//-----------------------------------------------------------------------------------

void draw_s(void)
{   
   int t;
   double unifrandoms[TTTT+2];
   double punnorm1, punnorm2, punnorm3, punnorm4, punnorm5, punnorm6, punnorm7, ss;
   
   for (t=1; t<=TTTT; t++) 
      YAST[t] = log((Yest[t]-impactX(t))*(Yest[t]-impactX(t))) + log(LAMBDAest[t]);   
   
   for (t=1; t<=TTTT; t++) unifrandoms[t]=uniform(0.0000,1.0000); 
    
   for (t=1; t<=TTTT; t++)
   {
      punnorm1 = 0.00730*dnorm(YAST[t],Hest[t]-11.40039,2.40748);
      punnorm2 = 0.10556*dnorm(YAST[t],Hest[t]- 5.24321,1.61669);
      punnorm3 = 0.00002*dnorm(YAST[t],Hest[t]- 9.83726,2.27585);
      punnorm4 = 0.04395*dnorm(YAST[t],Hest[t]+ 1.50746,0.40908);
      punnorm5 = 0.34001*dnorm(YAST[t],Hest[t]- 0.65098,0.80006);
      punnorm6 = 0.24566*dnorm(YAST[t],Hest[t]+ 0.52478,0.58329);
      punnorm7 = 0.25750*dnorm(YAST[t],Hest[t]- 2.35859,1.12366);      
      
      ss = punnorm1+punnorm2+punnorm3+punnorm4+punnorm5+punnorm6+punnorm7;
                
      if (unifrandoms[t]>(punnorm1+punnorm2+punnorm3+punnorm4+punnorm5+punnorm6)/ss) 
         sest[t]=7;
      else
         if (unifrandoms[t]>(punnorm1+punnorm2+punnorm3+punnorm4+punnorm5)/ss) 
	    sest[t]=6;
         else
	    if (unifrandoms[t]>(punnorm1+punnorm2+punnorm3+punnorm4)/ss) sest[t]=5;
            else
	       if (unifrandoms[t]>(punnorm1+punnorm2+punnorm3)/ss) sest[t]=4;
               else
	          if (unifrandoms[t]>(punnorm1+punnorm2)/ss) sest[t]=3;
                  else
		     if (unifrandoms[t]>punnorm1/ss) sest[t]=2;
                     else sest[t]=1;
   }
}   


//-----------------------------------------------------------------------------------

double minimum(double x, double y)
{
   if (x<=y) return x;
   else      return y;
}

//-----------------------------------------------------------------------------------
